from test.ClientServerTest import ClientServerTest
from udsoncan import MemoryLocation, AddressAndLengthFormatIdentifier
from udsoncan.exceptions import *

# Note : 
# MemoryLocation object is unit tested in a separate file (test_helper_class). 
# As it is the only parameter to be passed, no need to push this test too far for nothing.

class TestWriteMemoryByAddress(ClientServerTest):

    def test_4byte_block(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x3D\x12\x12\x34\x04\x66\x77\x88\x99")
        self.conn.fromuserqueue.put(b"\x7D\x12\x12\x34\x04")

    def _test_4byte_block(self):
        memloc = MemoryLocation(address=0x1234, memorysize=4, address_format=16, memorysize_format=8)
        response = self.udsclient.write_memory_by_address(memloc, b'\x66\x77\x88\x99')

        self.assertEqual(response.service_data.alfid_echo, memloc.alfid.get_byte_as_int())
        self.assertEqual(response.service_data.memory_location_echo.address, 0x1234)
        self.assertEqual(response.service_data.memory_location_echo.memorysize, 4)

    def test_4byte_block_spr_no_effect(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x3D\x12\x12\x34\x04\x66\x77\x88\x99")
        self.conn.fromuserqueue.put(b"\x7D\x12\x12\x34\x04")

    def _test_4byte_block_spr_no_effect(self):
        memloc = MemoryLocation(address=0x1234, memorysize=4, address_format=16, memorysize_format=8)

        with self.udsclient.suppress_positive_response:
            response = self.udsclient.write_memory_by_address(memloc, b'\x66\x77\x88\x99')
            self.assertEqual(response.service_data.alfid_echo, memloc.alfid.get_byte_as_int())
            self.assertEqual(response.service_data.memory_location_echo.address, 0x1234)
            self.assertEqual(response.service_data.memory_location_echo.memorysize, 4)

    def test_config_format(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x3D\x24\x00\x00\x12\x34\x00\x04\x66\x77\x88\x99")
        self.conn.fromuserqueue.put(b"\x7D\x24\x00\x00\x12\x34\x00\x04")

    def _test_config_format(self):
        self.udsclient.config['server_address_format'] = 32
        self.udsclient.config['server_memorysize_format'] = 16
        response = self.udsclient.write_memory_by_address(MemoryLocation(address=0x1234, memorysize=4), b'\x66\x77\x88\x99')
        alfid = AddressAndLengthFormatIdentifier(address_format = self.udsclient.config['server_address_format'], memorysize_format=self.udsclient.config['server_memorysize_format'])
        self.assertEqual(response.service_data.alfid_echo, alfid.get_byte_as_int())
        self.assertEqual(response.service_data.memory_location_echo.address, 0x1234)
        self.assertEqual(response.service_data.memory_location_echo.memorysize, 4)

    def test_4byte_block_client_extra_bytes(self):
        self.wait_request_and_respond(b'\x7D\x12\x12\x34\x04\x01\x02\x03\x04')

    def _test_4byte_block_client_extra_bytes(self):
        memloc = MemoryLocation(address=0x1234, memorysize=4, address_format=16, memorysize_format=8)
        self.udsclient.write_memory_by_address(memloc, b'\x66\x77\x88\x99\xAA\xBB')

    def test_4byte_block_harmless_extra_bytes(self):
        self.wait_request_and_respond(b'\x7D\x12\x12\x34\x04\x01\x02\x03\x04\x05')

    def _test_4byte_block_harmless_extra_bytes(self):
        memloc = MemoryLocation(address=0x1234, memorysize=4, address_format=16, memorysize_format=8)
        self.udsclient.write_memory_by_address(memloc, b'\x66\x77\x88\x99')

    def test_bad_echo_ali_exception(self):
        self.wait_request_and_respond(b'\x7D\x21\x12\x34\x04')

    def _test_bad_echo_ali_exception(self):
        with self.assertRaises(UnexpectedResponseException):
            memloc = MemoryLocation(address=0x1234, memorysize=4, address_format=16, memorysize_format=8)
            self.udsclient.write_memory_by_address(memloc, b'\x66\x77\x88\x99')

    def test_bad_echo_ali_no_exception(self):
        self.wait_request_and_respond(b'\x7D\x21\x12\x34\x04')

    def _test_bad_echo_ali_no_exception(self):
        self.udsclient.config['exception_on_unexpected_response'] = False
        memloc = MemoryLocation(address=0x1234, memorysize=4, address_format=16, memorysize_format=8)
        response = self.udsclient.write_memory_by_address(memloc, b'\x66\x77\x88\x99')
        self.assertTrue(response.valid)	
        self.assertTrue(response.unexpected)

    def test_bad_echo_address_exception(self):
        self.wait_request_and_respond(b'\x7D\x12\x12\x35\x04')

    def _test_bad_echo_address_exception(self):
        with self.assertRaises(UnexpectedResponseException):
            memloc = MemoryLocation(address=0x1234, memorysize=4, address_format=16, memorysize_format=8)
            self.udsclient.write_memory_by_address(memloc, b'\x66\x77\x88\x99')

    def test_bad_echo_address_no_exception(self):
        self.wait_request_and_respond(b'\x7D\x12\x12\x35\x04')

    def _test_bad_echo_address_no_exception(self):
        self.udsclient.config['exception_on_unexpected_response'] = False		
        memloc = MemoryLocation(address=0x1234, memorysize=4, address_format=16, memorysize_format=8)
        response = self.udsclient.write_memory_by_address(memloc, b'\x66\x77\x88\x99')
        self.assertTrue(response.valid)		
        self.assertTrue(response.unexpected)

    def test_bad_echo_memorysize_exception(self):
        self.wait_request_and_respond(b'\x7D\x12\x12\x34\x05')

    def _test_bad_echo_memorysize_exception(self):
        with self.assertRaises(UnexpectedResponseException):
            memloc = MemoryLocation(address=0x1234, memorysize=4, address_format=16, memorysize_format=8)
            self.udsclient.write_memory_by_address(memloc, b'\x66\x77\x88\x99')

    def test_bad_echo_memorysize_no_exception(self):
        self.wait_request_and_respond(b'\x7D\x12\x12\x34\x05')

    def _test_bad_echo_memorysize_no_exception(self):
        self.udsclient.config['exception_on_unexpected_response'] = False			
        memloc = MemoryLocation(address=0x1234, memorysize=4, address_format=16, memorysize_format=8)
        response = self.udsclient.write_memory_by_address(memloc, b'\x66\x77\x88\x99')
        self.assertTrue(response.valid)
        self.assertTrue(response.unexpected)

    def test_request_denied_exception(self):
        self.wait_request_and_respond(b"\x7F\x3D\x45") #Request Out Of Range

    def _test_request_denied_exception(self):
        with self.assertRaises(NegativeResponseException) as handle:
            memloc = MemoryLocation(address=0x1234, memorysize=4, address_format=16, memorysize_format=8)
            self.udsclient.write_memory_by_address(memloc, b'\x66\x77\x88\x99')

    def test_request_denied_no_exception(self):
        self.wait_request_and_respond(b"\x7F\x3D\x45") #Request Out Of Range

    def _test_request_denied_no_exception(self):
        self.udsclient.config['exception_on_negative_response'] = False
        memloc = MemoryLocation(address=0x1234, memorysize=4, address_format=16, memorysize_format=8)
        response = self.udsclient.write_memory_by_address(memloc, b'\x66\x77\x88\x99')
        self.assertTrue(response.valid)
        self.assertFalse(response.positive)

    def test_request_invalid_service_exception(self):
        self.wait_request_and_respond(b"\x00\x12\x12\x34\x04") #Inexistent Service

    def _test_request_invalid_service_exception(self):
        with self.assertRaises(InvalidResponseException) as handle:
            memloc = MemoryLocation(address=0x1234, memorysize=4, address_format=16, memorysize_format=8)
            self.udsclient.write_memory_by_address(memloc, b'\x66\x77\x88\x99')

    def test_request_invalid_service_no_exception(self):
        self.wait_request_and_respond(b"\x00\x12\x12\x34\x04") #Inexistent Service

    def _test_request_invalid_service_no_exception(self):
        self.udsclient.config['exception_on_invalid_response'] = False
        memloc = MemoryLocation(address=0x1234, memorysize=4, address_format=16, memorysize_format=8)
        response = self.udsclient.write_memory_by_address(memloc, b'\x66\x77\x88\x99')
        self.assertFalse(response.valid)

    def test_wrong_service_exception(self):
        self.wait_request_and_respond(b"\x7E\x12\x12\x34\x04") # Valid but wrong service (Tester Present)

    def _test_wrong_service_exception(self):
        with self.assertRaises(UnexpectedResponseException) as handle:
            memloc = MemoryLocation(address=0x1234, memorysize=4, address_format=16, memorysize_format=8)
            self.udsclient.write_memory_by_address(memloc, b'\x66\x77\x88\x99')

    def test_wrong_service_no_exception(self):
        self.wait_request_and_respond(b"\x7E\x12\x12\x34\x04") # Valid but wrong service (Tester Present)

    def _test_wrong_service_no_exception(self):
        self.udsclient.config['exception_on_unexpected_response'] = False	
        memloc = MemoryLocation(address=0x1234, memorysize=4, address_format=16, memorysize_format=8)
        response = self.udsclient.write_memory_by_address(memloc, b'\x66\x77\x88\x99')
        self.assertTrue(response.valid)
        self.assertTrue(response.unexpected)		

    def test_bad_param(self):
        pass

    def _test_bad_param(self):
        memloc = MemoryLocation(address=0x1234, memorysize=4, address_format=16, memorysize_format=8)

        with self.assertRaises(ValueError):
            self.udsclient.write_memory_by_address(1, b'\x00\x01')

        with self.assertRaises(ValueError):
            self.udsclient.write_memory_by_address('aaa',  b'\x00\x01')

        with self.assertRaises(ValueError):
            self.udsclient.write_memory_by_address(memloc,  1)

        with self.assertRaises(ValueError):
            self.udsclient.write_memory_by_address(memloc,  'aaa')



from udsoncan.client import Client
from udsoncan import services
from udsoncan.exceptions import *

from test.ClientServerTest import ClientServerTest

class TestECUReset(ClientServerTest):
    def __init__(self, *args, **kwargs):
        ClientServerTest.__init__(self, *args, **kwargs)

    def test_ecu_reset_success(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x11\x55")
        self.conn.fromuserqueue.put(b"\x51\x55")	# Positive response

    def _test_ecu_reset_success(self):
        response = self.udsclient.ecu_reset(0x55)
        self.assertTrue(response.positive)
        self.assertEqual(response.service_data.reset_type_echo, 0x55)

    def test_ecu_reset_success_spr(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x11\xD5")
        self.conn.fromuserqueue.put("wait")	# Synchronize

    def _test_ecu_reset_success_spr(self):
        with self.udsclient.suppress_positive_response:
            response = self.udsclient.ecu_reset(0x55)
            self.assertEqual(response, None)
        self.conn.fromuserqueue.get(timeout=0.2)	#Avoid closing connection prematurely

    def test_ecu_reset_success_pdt(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x11\x04")
        self.conn.fromuserqueue.put(b"\x51\x04\x23")	# Positive response 

    def _test_ecu_reset_success_pdt(self):
        response = self.udsclient.ecu_reset(services.ECUReset.ResetType.enableRapidPowerShutDown)
        self.assertTrue(response.positive)
        self.assertEqual(response.service_data.reset_type_echo, services.ECUReset.ResetType.enableRapidPowerShutDown)
        self.assertEqual(response.service_data.powerdown_time, 0x23)

    def test_ecu_reset_denied_exception(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x11\x55")
        self.conn.fromuserqueue.put(b"\x7F\x11\x33") #Security Access Denied

    def _test_ecu_reset_denied_exception(self):
        with self.assertRaises(NegativeResponseException) as handle:
            self.udsclient.ecu_reset(0x55)
        response = handle.exception.response

        self.assertTrue(response.valid)
        self.assertTrue(issubclass(response.service, services.ECUReset))
        self.assertEqual(response.code, 0x33)

    def test_ecu_reset_denied_no_exception(self):
        self.wait_request_and_respond(b"\x7F\x11\x33") #Security Access Denied

    def _test_ecu_reset_denied_no_exception(self):
        self.udsclient.config['exception_on_negative_response'] = False
        response = self.udsclient.ecu_reset(0x55)

        self.assertTrue(response.valid)
        self.assertFalse(response.positive)
        self.assertTrue(issubclass(response.service, services.ECUReset))
        self.assertEqual(response.code, 0x33)

    def test_ecu_reset_invalidservice_exception(self):
        self.wait_request_and_respond(b"\x00\x55") #Inexistent Service

    def _test_ecu_reset_invalidservice_exception(self):
        with self.assertRaises(InvalidResponseException) as handle:
            self.udsclient.ecu_reset(0x55)

    def test_ecu_reset_invalidservice_no_exception(self):
        self.wait_request_and_respond(b"\x00\x55") #Inexistent Service

    def _test_ecu_reset_invalidservice_no_exception(self):
        self.udsclient.config['exception_on_invalid_response'] = False
        response = self.udsclient.ecu_reset(0x55)

    def test_ecu_reset_wrongservice_exception(self):
        self.wait_request_and_respond(b"\x7E\x00") # Valid but wrong service (Tester Present)

    def _test_ecu_reset_wrongservice_exception(self):
        with self.assertRaises(UnexpectedResponseException) as handle:
            self.udsclient.ecu_reset(0x55)

    def test_ecu_reset_wrongservice_no_exception(self):
        self.wait_request_and_respond(b"\x7E\x00") # Valid but wrong service (Tester Present)

    def _test_ecu_reset_wrongservice_no_exception(self):
        self.udsclient.config['exception_on_unexpected_response'] = False
        response = self.udsclient.ecu_reset(0x55)
        self.assertTrue(response.valid)
        self.assertTrue(response.unexpected)

    def test_ecu_reset_missing_pdt_exception(self):
        self.wait_request_and_respond(b"\x51\x04") # Valid but wrong service (Tester Present)

    def _test_ecu_reset_missing_pdt_exception(self):
        with self.assertRaises(InvalidResponseException) as handle:
            self.udsclient.ecu_reset(services.ECUReset.ResetType.enableRapidPowerShutDown)

    def test_ecu_reset_missing_pdt_no_exception(self):
        self.wait_request_and_respond(b"\x51\x04") # Valid but wrong service (Tester Present)

    def _test_ecu_reset_missing_pdt_no_exception(self):
        self.udsclient.config['exception_on_invalid_response'] = False
        response = self.udsclient.ecu_reset(services.ECUReset.ResetType.enableRapidPowerShutDown)
        self.assertFalse(response.valid)

    def test_bad_param(self):
        pass

    def _test_bad_param(self):
        with self.assertRaises(ValueError):
            response = self.udsclient.ecu_reset(0x100)

        with self.assertRaises(ValueError):
            response = self.udsclient.ecu_reset(-1)



from udsoncan import services
from udsoncan.exceptions import *

from test.ClientServerTest import ClientServerTest


class TestAccessTimingParameter(ClientServerTest):
    def __init__(self, *args, **kwargs):
        ClientServerTest.__init__(self, *args, **kwargs)

    def test_read_extended_params_success(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x83\x01")
        self.conn.fromuserqueue.put(b"\xC3\x01\x99\x88\x77\x66")  # Positive response

    def _test_read_extended_params_success(self):
        response = self.udsclient.read_extended_timing_parameters()
        self.assertEqual(response.service_data.access_type_echo, 0x01)
        self.assertEqual(response.service_data.timing_param_record, b"\x99\x88\x77\x66")

    def test_read_extended_params_success_spr(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x83\x81")
        self.conn.fromuserqueue.put('wait')

    def _test_read_extended_params_success_spr(self):
        with self.udsclient.suppress_positive_response:
            response = self.udsclient.read_extended_timing_parameters()
        self.assertEqual(response, None)
        self.conn.fromuserqueue.get(timeout=0.2)  # Avoid closing the conenction prematurely

    def test_read_active_params_success(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x83\x03")
        self.conn.fromuserqueue.put(b"\xC3\x03\x99\x88\x77\x66")  # Positive response

    def _test_read_active_params_success(self):
        response = self.udsclient.read_active_timing_parameters()
        self.assertTrue(response.positive)
        self.assertEqual(response.service_data.access_type_echo, 0x03)
        self.assertEqual(response.service_data.timing_param_record, b"\x99\x88\x77\x66")

    def test_set_params_success(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x83\x04\x11\x22\x33\x44")
        self.conn.fromuserqueue.put(b"\xC3\x04")  # Positive response

    def _test_set_params_success(self):
        self.udsclient.set_timing_parameters(params=b"\x11\x22\x33\x44")

    def test_set_params_success_default(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x83\x02")
        self.conn.fromuserqueue.put(b"\xC3\x02")  # Positive response

    def _test_set_params_success_default(self):
        self.udsclient.reset_default_timing_parameters()

    def test_set_params_denied_exception(self):
        self.wait_request_and_respond(b"\x7F\x83\x31")  # Request Out Of Range

    def _test_set_params_denied_exception(self):
        with self.assertRaises(NegativeResponseException) as handle:
            self.udsclient.access_timing_parameter(access_type=0x22)

    def test_set_params_denied_no_exception(self):
        self.wait_request_and_respond(b"\x7F\x83\x31")  # Request Out Of Range

    def _test_set_params_denied_no_exception(self):
        self.udsclient.config['exception_on_negative_response'] = False
        response = self.udsclient.access_timing_parameter(access_type=0x22)
        self.assertTrue(response.valid)
        self.assertFalse(response.positive)

    def test_set_params_invalidservice_exception(self):
        self.wait_request_and_respond(b"\x00\x22")  # Inexistent Service

    def _test_set_params_invalidservice_exception(self):
        with self.assertRaises(InvalidResponseException) as handle:
            self.udsclient.access_timing_parameter(access_type=0x22)

    def test_set_params_invalidservice_no_exception(self):
        self.wait_request_and_respond(b"\x00\x22")  # Inexistent Service

    def _test_set_params_invalidservice_no_exception(self):
        self.udsclient.config['exception_on_invalid_response'] = False
        response = self.udsclient.access_timing_parameter(access_type=0x22)
        self.assertFalse(response.valid)

    def test_routine_control_wrongservice_exception(self):
        self.wait_request_and_respond(b"\x7E\x22")  # Valid but wrong service (Tester Present)

    def _test_routine_control_wrongservice_exception(self):
        with self.assertRaises(UnexpectedResponseException) as handle:
            self.udsclient.access_timing_parameter(access_type=0x22)

    def test_routine_control_wrongservice_no_exception(self):
        self.wait_request_and_respond(b"\x7E\x22")  # Valid but wrong service (Tester Present)

    def _test_routine_control_wrongservice_no_exception(self):
        self.udsclient.config['exception_on_unexpected_response'] = False
        response = self.udsclient.access_timing_parameter(access_type=0x22)
        self.assertTrue(response.valid)
        self.assertTrue(response.unexpected)

    def test_access_timing_params_bad_access_type_exception(self):
        self.wait_request_and_respond(b"\xC3\x23")  # Valid but access type

    def _test_access_timing_params_bad_access_type_exception(self):
        with self.assertRaises(UnexpectedResponseException) as handle:
            self.udsclient.access_timing_parameter(access_type=0x22)

    def test_access_timing_params_bad_access_type_no_exception(self):
        self.wait_request_and_respond(b"\xC3\x23")  # Valid but access type

    def _test_access_timing_params_bad_access_type_no_exception(self):
        self.udsclient.config['exception_on_unexpected_response'] = False
        response = self.udsclient.access_timing_parameter(access_type=0x22)
        self.assertTrue(response.valid)
        self.assertTrue(response.unexpected)

    def test_bad_param(self):
        pass

    def _test_bad_param(self):
        with self.assertRaises(ValueError):
            response = self.udsclient.access_timing_parameter(access_type=-1)

        with self.assertRaises(ValueError):
            response = self.udsclient.access_timing_parameter(access_type=0x100)

        with self.assertRaises(ValueError):
            response = self.udsclient.access_timing_parameter(
                access_type=services.AccessTimingParameter.AccessType.setTimingParametersToGivenValues, timing_param_record=None)

        with self.assertRaises(ValueError):
            response = self.udsclient.access_timing_parameter(
                access_type=services.AccessTimingParameter.AccessType.readExtendedTimingParameterSet, timing_param_record=b"\xaa\xbb")

        with self.assertRaises(ValueError):
            response = self.udsclient.access_timing_parameter(
                access_type=services.AccessTimingParameter.AccessType.setTimingParametersToDefaultValues, timing_param_record=123)



from udsoncan.common.DidCodec import DidCodec
from typing import TypedDict, Dict, Optional, Any, Callable, Union, Type

SecurityAlgoType = Callable[[int, bytes, Any], bytes]


CodecDefinition = Union[str, DidCodec, Type[DidCodec]]
DIDConfig = Dict[int, CodecDefinition]


class IOConfigEntry(TypedDict, total=False):
    codec: CodecDefinition
    mask: Optional[Dict[str, int]]
    mask_size: Optional[int]


IOConfig = Dict[int, Union[IOConfigEntry, CodecDefinition]]


class ClientConfig(TypedDict, total=False):
    exception_on_negative_response: bool
    exception_on_invalid_response: bool
    exception_on_unexpected_response: bool
    security_algo: Optional[SecurityAlgoType]
    security_algo_params: Optional[Any]
    tolerate_zero_padding: bool
    ignore_all_zero_dtc: bool
    dtc_snapshot_did_size: int
    server_address_format: Optional[int]
    server_memorysize_format: Optional[int]
    data_identifiers: Dict[int, CodecDefinition]
    input_output: IOConfig
    request_timeout: float
    p2_timeout: float
    p2_star_timeout: float
    standard_version: int
    use_server_timing: bool
    logger_name: str
    extended_data_size: Optional[Union[int, Dict[int, int]]]



from udsoncan import Request, Response
from udsoncan.exceptions import *
from udsoncan.BaseService import BaseService, BaseSubfunction, BaseResponseData
from udsoncan.ResponseCode import ResponseCode
import udsoncan.tools as tools

from typing import Optional, cast


class AccessTimingParameter(BaseService):
    _sid = 0x83

    class AccessType(BaseSubfunction):
        """
        AccessTimingParameter defined subfunctions
        """

        __pretty_name__ = 'access type'

        readExtendedTimingParameterSet = 1
        setTimingParametersToDefaultValues = 2
        readCurrentlyActiveTimingParameters = 3
        setTimingParametersToGivenValues = 4

    class ResponseData(BaseResponseData):
        """
        .. data:: access_type_echo

                Request subfunction echoed back by the server

        .. data:: timing_param_record

                Additional data associated with the response.
        """

        access_type_echo: int
        timing_param_record: bytes

        def __init__(self, access_type_echo: int, timing_param_record: bytes):
            super().__init__(AccessTimingParameter)
            self.access_type_echo = access_type_echo
            self.timing_param_record = timing_param_record

    supported_negative_response = [ResponseCode.SubFunctionNotSupported,
                                   ResponseCode.IncorrectMessageLengthOrInvalidFormat,
                                   ResponseCode.ConditionsNotCorrect,
                                   ResponseCode.RequestOutOfRange
                                   ]

    class InterpretedResponse(Response):
        service_data: "AccessTimingParameter.ResponseData"

    @classmethod
    def make_request(cls, access_type: int, timing_param_record: Optional[bytes] = None) -> Request:
        """
        Generates a request for AccessTimingParameter

        :param access_type: Service subfunction. Allowed values are from 0 to 0x7F
        :type access_type: int

        :param timing_param_record: Data associated with request. Must be present only when access_type=``AccessType.setTimingParametersToGivenValues`` (4)
        :type timing_param_record: bytes

        :raises ValueError: If parameters are out of range, missing or wrong type
        """

        tools.validate_int(access_type, min=0, max=0x7F, name='Access type')

        if timing_param_record is not None and access_type != cls.AccessType.setTimingParametersToGivenValues:
            raise ValueError('timing_param_record can only be set when access_type is setTimingParametersToGivenValues"')

        if timing_param_record is None and access_type == cls.AccessType.setTimingParametersToGivenValues:
            raise ValueError('A timing_param_record must be provided when access_type is "setTimingParametersToGivenValues"')

        request = Request(service=cls, subfunction=access_type, data=bytes())
        assert request.data is not None

        if timing_param_record is not None:
            if not isinstance(timing_param_record, bytes):
                raise ValueError("timing_param_record must be a valid bytes objects")
            request.data += timing_param_record

        return request

    @classmethod
    def interpret_response(cls, response: Response) -> InterpretedResponse:
        """
        Populates the response ``service_data`` property with an instance of :class:`AccessTimingParameter.ResponseData<udsoncan.services.AccessTimingParameter.ResponseData>`

        :param response: The received response to interpret
        :type response: :ref:`Response<Response>`

        :raises InvalidResponseException: If length of ``response.data`` is too short
        """
        if response.data is None:
            raise InvalidResponseException(response, "No data in response")

        if len(response.data) < 1:
            raise InvalidResponseException(response, "Response data must be at least 1 byte")

        response.service_data = cls.ResponseData(
            access_type_echo=response.data[0],
            timing_param_record=response.data[1:] if len(response.data) > 1 else b''
        )
        return cast(AccessTimingParameter.InterpretedResponse, response)



from udsoncan import Response
from udsoncan.BaseService import BaseService
from test.UdsTest import UdsTest
import inspect

class DummyServiceNormal(BaseService):
    _sid = 0x13

    def subfunction_id(self):
        return 0x44

class DummyServiceNoSubunction(BaseService):
    _sid = 0x13
    _use_subfunction = False

class DummyServiceNoResponseData(BaseService):
    _sid = 0x13
    _no_response_data = True

class RandomClass:
    pass


class TestResponse(UdsTest):

    def test_create_from_instance_ok(self):
        response = Response(DummyServiceNormal(), code = 0x22)
        self.assertTrue(response.valid)
        self.assertEqual(response.service.request_id(), 0x13)
        self.assertEqual(response.code, 0x22)

    def test_create_from_class_ok(self):
        response = Response(DummyServiceNormal, code=0x22)
        self.assertTrue(response.valid)
        self.assertEqual(response.service.request_id(), 0x13)
        self.assertEqual(response.code, 0x22)

    def test_make_payload_basic_positive(self):
        response = Response(DummyServiceNormal(), code = 0, data=b"\x01")
        self.assertTrue(response.positive)
        self.assertTrue(response.valid)
        payload = response.get_payload()
        self.assertEqual(b"\x53\x01", payload)	# Original ID + 0x40

    def test_make_payload_basic_negative(self):
        response = Response(DummyServiceNormal(), code = 0x10) # General Reject
        self.assertFalse(response.positive)
        self.assertTrue(response.valid)
        payload = response.get_payload()
        self.assertEqual(b"\x7F\x13\x10", payload)	# Original ID + 0x40. 7F indicate negative

    def test_make_payload_custom_data_negative(self):
        response = Response(DummyServiceNormal(), code = 0x10)
        self.assertTrue(response.valid)
        self.assertFalse(response.positive)
        response.data = b"\x12\x34\x56\x78"
        payload = response.get_payload()
        self.assertEqual(b"\x7F\x13\x10\x12\x34\x56\x78", payload)

    def test_from_payload_basic_positive(self):
        payload=b'\x7E\x00'	# 0x7e = TesterPresent
        response = Response.from_payload(payload)
        self.assertTrue(response.valid)
        self.assertTrue(response.positive)
        self.assertEqual(response.service.response_id(), 0x7E)
        self.assertEqual(response.code, 0)

    def test_from_payload_basic_negative(self):
        payload=b'\x7F\x3E\x10'	# 0x3e = TesterPresent, 0x10 = General Reject
        response = Response.from_payload(payload)
        self.assertTrue(response.valid)
        self.assertFalse(response.positive)
        self.assertEqual(response.service.response_id(), 0x7E)
        self.assertEqual(response.code, 0x10)		

    def test_from_payload_custom_data_positive(self):
        payload=b'\x7E\x01\x12\x34\x56\x78'	# 0x3E = TesterPresent
        response = Response.from_payload(payload)
        self.assertTrue(response.valid)
        self.assertTrue(response.positive)
        self.assertEqual(response.service.response_id(), 0x7E)	
        self.assertEqual(response.data, b'\x01\x12\x34\x56\x78')

    def test_from_payload_custom_data_negative(self):
        payload=b'\x7F\x3E\x10\x12\x34\x56\x78'	# 0x3E = TesterPresent, 0x10 = General Reject
        response = Response.from_payload(payload)
        self.assertTrue(response.valid)
        self.assertEqual(response.service.response_id(), 0x7E)
        self.assertEqual(response.code, 0x10)	
        self.assertEqual(response.data, b'\x12\x34\x56\x78')

    def test_from_empty_payload(self):
        payload = b''
        response = Response.from_payload(payload)
        self.assertFalse(response.valid)
        self.assertIsNone(response.service)
        self.assertEqual(b'', response.data)

    def test_from_bad_payload(self):
        payload = b'\xFF\xFF'
        response = Response.from_payload(payload)
        self.assertFalse(response.valid)
        self.assertIsNone(response.service)
        self.assertEqual(b'', response.data)

    def test_str_repr(self):
        response = Response(DummyServiceNormal, code=0x22)
        str(response)
        response.__repr__()

    def test_from_input_param(self):
        with self.assertRaises(ValueError):
            response = Response(service = "a string")	

        with self.assertRaises(ValueError):
            response = Response(service = RandomClass())	

        with self.assertRaises(ValueError):
            response = Response(service=DummyServiceNormal(), code = "string")	

        with self.assertRaises(ValueError):
            response = Response(service=DummyServiceNormal(), code = -1)	

        with self.assertRaises(ValueError):
            response = Response(service=DummyServiceNormal(), code = 0x100)

        with self.assertRaises(ValueError):
            response = Response(service=DummyServiceNormal(), code = 0x10, data=11)	

    def test_all_response_code_have_version(self):
        codes = [ member[1] for member in inspect.getmembers(Response.Code) if isinstance(member[1], int) and not member[0].startswith('_')]
        self.assertGreater(len(codes), 0)
        for code in codes:
            # Make sure no exception is raised
            Response.Code.is_supported_by_standard(code, 2006)
            Response.Code.is_supported_by_standard(code, 2013)
            Response.Code.is_supported_by_standard(code, 2020)

        # Case of known code introduced in 2020
        self.assertFalse(Response.Code.is_supported_by_standard(Response.Code.ResourceTemporarilyNotAvailable , 2006))
        self.assertFalse(Response.Code.is_supported_by_standard(Response.Code.ResourceTemporarilyNotAvailable , 2013))
        self.assertTrue(Response.Code.is_supported_by_standard(Response.Code.ResourceTemporarilyNotAvailable , 2020))

        self.assertTrue(Response.Code.is_supported_by_standard(Response.Code.GeneralReject , 2006))
        self.assertTrue(Response.Code.is_supported_by_standard(Response.Code.GeneralReject , 2013))
        self.assertTrue(Response.Code.is_supported_by_standard(Response.Code.GeneralReject , 2020))


from udsoncan import services
from udsoncan.exceptions import *

from test.ClientServerTest import ClientServerTest

class TestRequestSeed(ClientServerTest):
    def __init__(self, *args, **kwargs):
        ClientServerTest.__init__(self, *args, **kwargs)

    def test_request_seed_success(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x27\x05")
        self.conn.fromuserqueue.put(b"\x67\x05\x99\x88\x77\x66")	# Positive response

    def _test_request_seed_success(self):
        response = self.udsclient.request_seed(0x05)
        self.assertEqual(response.service_data.seed, b"\x99\x88\x77\x66")

    def test_request_seed_success_spr(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x27\x85")
        self.conn.fromuserqueue.put('wait')	# Synchronize 

    def _test_request_seed_success_spr(self):
        with self.udsclient.suppress_positive_response:
            response = self.udsclient.request_seed(0x05)
            self.assertEqual(response, None)
        self.conn.fromuserqueue.get(timeout=0.2)	#Avoid closing connection prematurely

    def test_request_seed_denied_exception(self):
        self.wait_request_and_respond(b"\x7F\x27\x22")	# Conditions Not Correct

    def _test_request_seed_denied_exception(self):
        with self.assertRaises(NegativeResponseException) as handle:
            self.udsclient.request_seed(0x05)
        response = handle.exception.response

        self.assertTrue(response.valid)
        self.assertFalse(response.positive)
        self.assertTrue(issubclass(response.service, services.SecurityAccess))
        self.assertEqual(response.code, 0x22)

    def test_request_seed_denied_no_exception(self):
        self.wait_request_and_respond(b"\x7F\x27\x22")	# Conditions Not Correct

    def _test_request_seed_denied_no_exception(self):
        self.udsclient.config['exception_on_negative_response'] = False
        response = self.udsclient.request_seed(0x05)

        self.assertTrue(response.valid)
        self.assertFalse(response.positive)
        self.assertTrue(issubclass(response.service, services.SecurityAccess))
        self.assertEqual(response.code, 0x22)

    def test_request_seed_bad_subfn_exception(self):
        self.wait_request_and_respond(b"\x67\x06\x99\x88\x77\x66")	# Positive response with wrong subfunction

    def _test_request_seed_bad_subfn_exception(self):
        with self.assertRaises(UnexpectedResponseException) as handle:
            self.udsclient.request_seed(0x05)

    def test_request_seed_bad_subfn_no_exception(self):
        self.wait_request_and_respond(b"\x67\x06\x99\x88\x77\x66")	# Positive response with wrong subfunction

    def _test_request_seed_bad_subfn_no_exception(self):
        self.udsclient.config['exception_on_unexpected_response'] = False
        response = self.udsclient.request_seed(0x05)
        self.assertTrue(response.valid)
        self.assertTrue(response.unexpected)

    def test_request_seed_incomplete_response_exception(self):
        self.wait_request_and_respond(b"\x67\x05")	# Positive response with no seed

    def _test_request_seed_incomplete_response_exception(self):
        with self.assertRaises(InvalidResponseException) as handle:
            self.udsclient.request_seed(0x05)

    def test_request_seed_incomplete_response_no_exception(self):
        self.wait_request_and_respond(b"\x67\x05")	# Positive response with no seed

    def _test_request_seed_incomplete_response_no_exception(self):
        self.udsclient.config['exception_on_invalid_response'] = False
        response = self.udsclient.request_seed(0x05)
        self.assertFalse(response.valid)

    def test_request_seed_invalidservice_exception(self):
        self.wait_request_and_respond(b"\x00\x05") #Inexistent Service

    def _test_request_seed_invalidservice_exception(self):
        with self.assertRaises(InvalidResponseException) as handle:
            self.udsclient.request_seed(0x05)

    def test_request_seed_invalidservice_no_exception(self):
        self.wait_request_and_respond(b"\x00\x05") #Inexistent Service

    def _test_request_seed_invalidservice_no_exception(self):
        self.udsclient.config['exception_on_invalid_response'] = False
        response = self.udsclient.request_seed(0x05)
        self.assertFalse(response.valid)

    def test_request_seed_wrongservice_exception(self):
        self.wait_request_and_respond(b"\x7E\x00") # Valid but wrong service (Tester Present)

    def _test_request_seed_wrongservice_exception(self):
        with self.assertRaises(UnexpectedResponseException) as handle:
            self.udsclient.request_seed(0x05)

    def test_request_seed_wrongservice_no_exception(self):
        self.wait_request_and_respond(b"\x7E\x00") # Valid but wrong service (Tester Present)

    def _test_request_seed_wrongservice_no_exception(self):
        self.udsclient.config['exception_on_unexpected_response'] = False
        response = self.udsclient.request_seed(0x05)
        self.assertTrue(response.valid)
        self.assertTrue(response.unexpected)

    def test_request_seed_bad_param(self):
        pass

    def _test_request_seed_bad_param(self):
        with self.assertRaises(ValueError):
            self.udsclient.request_seed(0x80)

        with self.assertRaises(ValueError):
            self.udsclient.request_seed(-1)


class TestSendKey(ClientServerTest):
    def __init__(self, *args, **kwargs):
        ClientServerTest.__init__(self, *args, **kwargs)

    def test_send_key_success(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x27\x06\x11\x22\x33\x44")
        self.conn.fromuserqueue.put(b"\x67\x06")	# Positive response

    def _test_send_key_success(self):
        response = self.udsclient.send_key(0x06,b"\x11\x22\x33\x44")
        self.assertTrue(response.positive)

    def test_send_key_denied_exception(self):
        self.wait_request_and_respond(b"\x7F\x27\x35")	# InvalidKey

    def _test_send_key_denied_exception(self):
        with self.assertRaises(NegativeResponseException) as handle:
            self.udsclient.send_key(0x06, b"\x11\x22\x33\x44")
        response = handle.exception.response

        self.assertTrue(response.valid)
        self.assertFalse(response.positive)
        self.assertTrue(issubclass(response.service, services.SecurityAccess))
        self.assertEqual(response.code, 0x35)

    def test_send_key_denied_no_exception(self):
        self.wait_request_and_respond(b"\x7F\x27\x35")	# InvalidKey

    def _test_send_key_denied_no_exception(self):
        self.udsclient.config['exception_on_negative_response'] = False
        response = self.udsclient.send_key(0x06, b"\x11\x22\x33\x44")

        self.assertTrue(response.valid)
        self.assertFalse(response.positive)
        self.assertTrue(issubclass(response.service, services.SecurityAccess))
        self.assertEqual(response.code, 0x35)

    def test_send_key_bad_subfn_exception(self):
        self.wait_request_and_respond(b"\x67\x08\x99\x88\x77\x66")	# Positive response with wrong subfunction

    def _test_send_key_bad_subfn_exception(self):
        with self.assertRaises(UnexpectedResponseException) as handle:
            self.udsclient.send_key(0x06, b"\x11\x22\x33\x44")

    def test_send_key_bad_subfn_no_exception(self):
        self.wait_request_and_respond(b"\x67\x08\x99\x88\x77\x66")	# Positive response with wrong subfunction

    def _test_send_key_bad_subfn_no_exception(self):
        self.udsclient.config['exception_on_unexpected_response'] = False
        response = self.udsclient.send_key(0x06, b"\x11\x22\x33\x44")
        self.assertTrue(response.valid)
        self.assertTrue(response.unexpected)

    def test_send_key_invalidservice_exception(self):
        self.wait_request_and_respond(b"\x00\x06") #Inexistent Service

    def _test_send_key_invalidservice_exception(self):
        with self.assertRaises(InvalidResponseException) as handle:
            self.udsclient.send_key(0x06, b"\x11\x22\x33\x44")

    def test_send_key_invalidservice_no_exception(self):
        self.wait_request_and_respond(b"\x00\x06") #Inexistent Service

    def _test_send_key_invalidservice_no_exception(self):
        self.udsclient.config['exception_on_invalid_response'] = False
        response = self.udsclient.send_key(0x06, b"\x11\x22\x33\x44")
        self.assertFalse(response.valid)

    def test_send_key_wrongservice_exception(self):
        self.wait_request_and_respond(b"\x7E\x00") # Valid but wrong service (Tester Present)

    def _test_send_key_wrongservice_exception(self):
        with self.assertRaises(UnexpectedResponseException) as handle:
            self.udsclient.send_key(0x06, b"\x11\x22\x33\x44")

    def test_send_key_wrongservice_no_exception(self):
        self.wait_request_and_respond(b"\x7E\x00") # Valid but wrong service (Tester Present)

    def _test_send_key_wrongservice_no_exception(self):
        self.udsclient.config['exception_on_unexpected_response'] = False
        response = self.udsclient.send_key(0x06, b"\x11\x22\x33\x44")
        self.assertTrue(response.valid)
        self.assertTrue(response.unexpected)

    def test_send_key_bad_param(self):
        pass

    def _test_send_key_bad_param(self):
        with self.assertRaises(ValueError):
            self.udsclient.send_key(0x80, b"\x11\x22\x33\x44")

        with self.assertRaises(ValueError):
            self.udsclient.send_key(-1, b"\x11\x22\x33\x44")

        with self.assertRaises(ValueError):
            self.udsclient.send_key(1, 1)

        with self.assertRaises(ValueError):
            self.udsclient.send_key(1, 'xxx')


class TestUnlockSecurityService(ClientServerTest):
    def __init__(self, *args, **kwargs):
        ClientServerTest.__init__(self, *args, **kwargs)

    def dummy_algo_backward_compatible(self, seed, params=None):
        key = bytearray(seed)
        for i in range(len(key)):
            key[i] ^= params
        return bytes(key)

    def dummy_algo(self, level, seed, params=None):
        key = bytearray(seed)
        for i in range(len(key)):
            key[i] = (params + level + i + key[i])
        return bytes(key)

    def test_unlock_success(self):
        for i in range(2):
            request = self.conn.touserqueue.get(timeout=0.2)
            self.assertEqual(request, b"\x27\x07")      # Request seed
            self.conn.fromuserqueue.put(b"\x67\x07\x11\x22\x33\x44")    # Positive response
            request = self.conn.touserqueue.get(timeout=0.2)
            key = bytearray([(0x10 + 0x07+i + 0 + 0x11), (0x10 + 0x07+i + 1 + 0x22), (0x10 + 0x07+i + 2 + 0x33), (0x10 + 0x07+i + 3 + 0x44)])
            self.assertEqual(request, b"\x27\x08" + bytes(key))
            self.conn.fromuserqueue.put(b"\x67\x08")    # Positive response

    def _test_unlock_success(self):
        self.udsclient.config['security_algo'] = self.dummy_algo
        self.udsclient.config['security_algo_params'] = 0x10
        response = self.udsclient.unlock_security_access(0x07)  
        response = self.udsclient.unlock_security_access(0x08)  
        self.assertTrue(response.positive)

    def test_unlock_already_unlocked(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x67\x07\x00\x00\x00\x00")    # Positive response, seed=0 : already unlocked

    def _test_unlock_already_unlocked(self):
        self.udsclient.config['security_algo'] = self.dummy_algo
        self.udsclient.config['security_algo_params'] = 0x10
        response = self.udsclient.unlock_security_access(0x07)  
        self.assertTrue(response.positive)

    def test_unlock_success_backward_compatibility(self):
        for i in range(2):
            request = self.conn.touserqueue.get(timeout=0.2)
            self.assertEqual(request, b"\x27\x07")      # Request seed
            self.conn.fromuserqueue.put(b"\x67\x07\x11\x22\x33\x44")    # Positive response
            request = self.conn.touserqueue.get(timeout=0.2)
            self.assertEqual(request, b"\x27\x08\xEE\xDD\xCC\xBB")
            self.conn.fromuserqueue.put(b"\x67\x08")    # Positive response

    def _test_unlock_success_backward_compatibility(self):
        self.udsclient.config['security_algo'] = self.dummy_algo_backward_compatible
        self.udsclient.config['security_algo_params'] = 0xFF
        response = self.udsclient.unlock_security_access(0x07)  
        response = self.udsclient.unlock_security_access(0x08)  
        self.assertTrue(response.positive)


    def test_unlock_seed_fail_exception(self):
        self.wait_request_and_respond(b"\x7F\x27\x11")  

    def _test_unlock_seed_fail_exception(self):
        self.udsclient.config['security_algo'] = self.dummy_algo
        self.udsclient.config['security_algo_params'] = 0xFF
        with self.assertRaises(NegativeResponseException):
            response = self.udsclient.unlock_security_access(0x07)  

    def test_unlock_seed_fail_no_exception(self):
        self.wait_request_and_respond(b"\x7F\x27\x11")  

    def _test_unlock_seed_fail_no_exception(self):
        self.udsclient.config['security_algo'] = self.dummy_algo
        self.udsclient.config['security_algo_params'] = 0xFF
        self.udsclient.config['exception_on_negative_response'] = False
        response = self.udsclient.unlock_security_access(0x07)
        self.assertTrue(response.valid)
        self.assertFalse(response.positive)

    def test_no_algo_set(self):
        pass

    def _test_no_algo_set(self):
        with self.assertRaises(NotImplementedError):
            self.udsclient.unlock_security_access(0x07)



from udsoncan.client import Client
from udsoncan import services
from udsoncan.exceptions import *
from udsoncan import DidCodec
import struct

from test.ClientServerTest import ClientServerTest


class StubbedDidCodec(DidCodec):
    def encode(self, did_value):
        return struct.pack('B', did_value+1)

    def decode(self, did_payload):
        return struct.unpack('B', did_payload)[0] - 1

    def __len__(self):
        return 1

class StubbedCodecThatExpectTuple(DidCodec):
    def encode(self, the_tuple):
        if not isinstance(the_tuple, tuple):
            raise ValueError('Given value is not a tuple')
        return struct.pack('BB', the_tuple[0], the_tuple[1])

    def decode(self, did_payload):
        return struct.unpack('BB', did_payload)

    def __len__(self):
        return 2

class TestReadDataByIdentifier(ClientServerTest):
    def __init__(self, *args, **kwargs):
        ClientServerTest.__init__(self, *args, **kwargs)

    def postClientSetUp(self):
        self.udsclient.config["data_identifiers"] = {
                1 : '>H',
                2 : '<H',
                3 : StubbedDidCodec,
                4 : '<HHH',
                5 : StubbedCodecThatExpectTuple
        }

    def test_wdbi_single_success1(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x2E\x00\x01\x12\x34")
        self.conn.fromuserqueue.put(b"\x6E\x00\x01")	# Positive response

    def _test_wdbi_single_success1(self):
        response = self.udsclient.write_data_by_identifier(did = 1, value=0x1234)
        self.assertTrue(response.positive)

    def test_wdbi_single_success1_spr_no_effect(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x2E\x00\x01\x12\x34")
        self.conn.fromuserqueue.put(b"\x6E\x00\x01")	# Positive response

    def _test_wdbi_single_success1_spr_no_effect(self):
        with self.udsclient.suppress_positive_response:
            response = self.udsclient.write_data_by_identifier(did = 1, value=0x1234)
            self.assertTrue(response.positive)

    def test_wdbi_single_success2(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x2E\x00\x02\x34\x12")
        self.conn.fromuserqueue.put(b"\x6E\x00\x02")    # Positive response

    def _test_wdbi_single_success2(self):
        self.udsclient.write_data_by_identifier(did = 2, value=0x1234)

    #Test for issue #29
    def test_wdbi_single_success_multiple_vals_default_codec(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x2E\x00\x04\x22\x11\x44\x33\x66\x55")
        self.conn.fromuserqueue.put(b"\x6E\x00\x04")    # Positive response

    def _test_wdbi_single_success_multiple_vals_default_codec(self):
        self.udsclient.write_data_by_identifier(did = 4, value=(0x1122, 0x3344, 0x5566))

     #Test for issue #29
    def test_wdbi_codec_using_tuple(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x2E\x00\x05\x12\x34")
        self.conn.fromuserqueue.put(b"\x6E\x00\x05")    # Positive response

    def _test_wdbi_codec_using_tuple(self):
        self.udsclient.write_data_by_identifier(did = 5, value=(0x12, 0x34))

    def test_wdbi_incomplete_response_exception(self):
        self.wait_request_and_respond(b"\x6E\x00")	#Incomplete response

    def _test_wdbi_incomplete_response_exception(self):
        with self.assertRaises(InvalidResponseException):
            self.udsclient.write_data_by_identifier(did = 1, value=0x1234)

    def test_wdbi_incomplete_response_no_exception(self):
        self.wait_request_and_respond(b"\x6E\x00")	#Incomplete response

    def _test_wdbi_incomplete_response_no_exception(self):
        self.udsclient.config['exception_on_invalid_response'] = False
        response = self.udsclient.write_data_by_identifier(did = 1, value=0x1234)
        self.assertFalse(response.valid)

    def test_wdbi_unknown_did_exception(self):
        self.wait_request_and_respond(b"\x6E\x00\x09")	# Positive response

    def _test_wdbi_unknown_did_exception(self):
        with self.assertRaises(UnexpectedResponseException):
            self.udsclient.write_data_by_identifier(did = 1, value=0x1234)			

    def test_wdbi_unknown_did_no_exception(self):
        self.wait_request_and_respond(b"\x6E\x00\x09")	# Positive response

    def _test_wdbi_unknown_did_no_exception(self):
        self.udsclient.config['exception_on_unexpected_response'] = False
        response = self.udsclient.write_data_by_identifier(did = 1, value=0x1234)
        self.assertTrue(response.valid)
        self.assertTrue(response.unexpected)			

    def test_wdbi_unwanted_did_exception(self):
        self.wait_request_and_respond(b"\x6E\x00\x02")	# Positive response

    def _test_wdbi_unwanted_did_exception(self):
        with self.assertRaises(UnexpectedResponseException):
            self.udsclient.write_data_by_identifier(did = 1, value=0x1234)			

    def test_wdbi_unwanted_did_no_exception(self):
        self.wait_request_and_respond(b"\x6E\x00\x02")	# Positive response

    def _test_wdbi_unwanted_did_no_exception(self):
        self.udsclient.config['exception_on_unexpected_response'] = False
        response = self.udsclient.write_data_by_identifier(did = 1, value=0x1234)
        self.assertTrue(response.valid)
        self.assertTrue(response.unexpected)

    def test_wdbi_invalidservice_exception(self):
        self.wait_request_and_respond(b"\x00\x00\x01")	# Service is inexistant

    def _test_wdbi_invalidservice_exception(self):
        with self.assertRaises(InvalidResponseException) as handle:
            self.udsclient.write_data_by_identifier(did=1, value=0x1234)

    def test_wdbi_invalidservice_no_exception(self):
        self.wait_request_and_respond(b"\x00\x00\x01")	# Service is inexistant

    def _test_wdbi_invalidservice_no_exception(self):
        self.udsclient.config['exception_on_invalid_response'] = False
        response = self.udsclient.write_data_by_identifier(did=1, value=0x1234)
        self.assertFalse(response.valid)

    def test_wdbi_wrongservice_exception(self):
        self.wait_request_and_respond(b"\x50\x00\x01")	# Valid service, but not the one requested

    def _test_wdbi_wrongservice_exception(self):
        with self.assertRaises(UnexpectedResponseException) as handle:
            self.udsclient.write_data_by_identifier(did=1, value=0x1234)

    def test_wdbi_wrongservice_no_exception(self):
        self.wait_request_and_respond(b"\x50\x00\x01")	# Valid service, but not the one requested

    def _test_wdbi_wrongservice_no_exception(self):
        self.udsclient.config['exception_on_unexpected_response'] = False
        response = self.udsclient.write_data_by_identifier(did = 1, value=0x1234)
        self.assertTrue(response.valid)
        self.assertTrue(response.unexpected)

    def test_no_config(self):
        pass

    def _test_no_config(self):
        with self.assertRaises(ConfigError):
            self.udsclient.write_data_by_identifier(did = 6, value=0x1234) 



from udsoncan.exceptions import *

from test.ClientServerTest import ClientServerTest
from udsoncan import latest_standard


class TestClearDtc(ClientServerTest):
    def __init__(self, *args, **kwargs):
        ClientServerTest.__init__(self, *args, **kwargs)

    def test_clear_dtc_success(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x14\x12\x34\x56")
        self.conn.fromuserqueue.put(b"\x54")  # Positive response

    def _test_clear_dtc_success(self):
        response = self.udsclient.clear_dtc(0x123456)
        self.assertTrue(response.valid)
        self.assertTrue(response.positive)

    def test_clear_dtc_spr_no_effect(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x14\x12\x34\x56")
        self.conn.fromuserqueue.put(b"\x54")  # Positive response

    def _test_clear_dtc_spr_no_effect(self):
        with self.udsclient.suppress_positive_response:
            response = self.udsclient.clear_dtc(0x123456)
            self.assertTrue(response.valid)
            self.assertTrue(response.positive)

    def test_clear_dtc_all(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x14\xFF\xFF\xFF")
        self.conn.fromuserqueue.put(b"\x54")    # Positive response

    def _test_clear_dtc_all(self):
        response = self.udsclient.clear_dtc()
        self.assertTrue(response.valid)
        self.assertTrue(response.positive)

    def test_clear_dtc_with_memory_selection(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x14\x12\x34\x56\x99")
        self.conn.fromuserqueue.put(b"\x54")    # Positive response

    def _test_clear_dtc_with_memory_selection(self):
        response = self.udsclient.clear_dtc(group=0x123456, memory_selection=0x99)
        self.assertTrue(response.valid)
        self.assertTrue(response.positive)

    def test_clear_dtc_denied_exception(self):
        self.wait_request_and_respond(b"\x7F\x14\x31")  # Request Out Of Range

    def _test_clear_dtc_denied_exception(self):
        with self.assertRaises(NegativeResponseException):
            self.udsclient.clear_dtc(0x123456)

    def test_clear_dtc_denied_no_exception(self):
        self.wait_request_and_respond(b"\x7F\x14\x31")  # Request Out Of Range

    def _test_clear_dtc_denied_no_exception(self):
        self.udsclient.config['exception_on_negative_response'] = False
        response = self.udsclient.clear_dtc(0x123456)
        self.assertTrue(response.valid)
        self.assertFalse(response.positive)
        self.assertEqual(response.code, 0x31)

    def test_clear_dtc_invalidservice_exception(self):
        self.wait_request_and_respond(b"\x00")  # Inexistent Service

    def _test_clear_dtc_invalidservice_exception(self):
        with self.assertRaises(InvalidResponseException):
            self.udsclient.clear_dtc(0x123456)

    def test_clear_dtc_invalidservice_no_exception(self):
        self.wait_request_and_respond(b"\x00")  # Inexistent Service

    def _test_clear_dtc_invalidservice_no_exception(self):
        self.udsclient.config['exception_on_invalid_response'] = False
        response = self.udsclient.clear_dtc(0x123456)
        self.assertFalse(response.valid)

    def test_clear_dtc_wrongservice_exception(self):
        self.wait_request_and_respond(b"\x7E\x00")  # Valid but wrong service (Tester Present)

    def _test_clear_dtc_wrongservice_exception(self):
        with self.assertRaises(UnexpectedResponseException):
            self.udsclient.clear_dtc(0x123456)

    def test_clear_dtc_wrongservice_no_exception(self):
        self.wait_request_and_respond(b"\x7E\x00")  # Valid but wrong service (Tester Present)

    def _test_clear_dtc_wrongservice_no_exception(self):
        self.udsclient.config['exception_on_unexpected_response'] = False
        response = self.udsclient.clear_dtc(0x123456)
        self.assertTrue(response.valid)
        self.assertTrue(response.unexpected)

    def test_bad_param(self):
        pass

    def _test_bad_param(self):
        with self.assertRaises(ValueError):
            self.udsclient.clear_dtc(0x1000000)

        with self.assertRaises(ValueError):
            self.udsclient.clear_dtc(-1)

        with self.assertRaises(ValueError):
            self.udsclient.clear_dtc(1, memory_selection=-1)

        with self.assertRaises(ValueError):
            self.udsclient.clear_dtc(1, memory_selection=0x100)

        with self.assertRaises(NotImplementedError):
            self.udsclient.set_config('standard_version', 2013)
            self.udsclient.clear_dtc(1, memory_selection=0x100)
        self.udsclient.set_config('standard_version', latest_standard)



__all__ = [
    'TimeoutException',
    'NegativeResponseException',
    'InvalidResponseException',
    'UnexpectedResponseException',
    'ConfigError'
]
from udsoncan.Response import Response
from typing import Any


class TimeoutException(Exception):
    """
    Simple extension of ``Exception`` with no additional property. Raised when a timeout in the communication happens.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class NegativeResponseException(Exception):
    """
    Raised when the server returns a negative response (response code starting by 0x7F).
    The response that triggered the exception is available in ``e.response``

    :param response: The response that triggered the exception
    :type response: :ref:`Response<Response>`
    """
    response: Response

    def __init__(self, response: Response, *args, **kwargs):
        self.response = response
        msg = self.make_msg(response)
        if len(args) > 0:
            msg += " " + str(args[0])
            args = tuple(list(args)[1:])
        super().__init__(msg, *args, **kwargs)

    def make_msg(self, response: Response):
        servicename = response.service.get_name() + " " if response.service is not None else ""
        codestr = ""
        if response.code is not None:
            codestr = " (0x%x)" % response.code
        return "%sservice execution returned a negative response %s%s" % (servicename, response.code_name, codestr)


class InvalidResponseException(Exception):
    """
    Raised when a service fails to decode a server response data. A bad message length or a value that is out of range may both be valid causes.
    The response that triggered the exception is available in ``e.response``

    :param response: The response that triggered the exception
    :type response: :ref:`Response<Response>`
    """
    response: Response

    def __init__(self, response, *args, **kwargs):
        self.response = response
        msg = self.make_msg(response)
        if len(args) > 0:
            msg += " " + str(args[0])
            args = tuple(list(args)[1:])
        super().__init__(msg, *args, **kwargs)

    def make_msg(self, response: Response):
        servicename = response.service.get_name() + " " if response.service is not None else ""
        reason = "" if response.valid else " Reason : %s" % (response.invalid_reason)
        return "%sservice execution returned an invalid response.%s" % (servicename, reason)


class UnexpectedResponseException(Exception):
    """
    Raised when the client receives a valid response but considers the one received to not be the expected response.
    The response that triggered the exception is available in ``e.response``

    :param response: The response that triggered the exception
    :type response: :ref:`Response<Response>`

    :param details: Additional details about the error
    :type details: string
    """
    response: Response

    def __init__(self, response: Response, details="<No details given>", *args, **kwargs):
        self.response = response
        msg = self.make_msg(response, details)
        if len(args) > 0:
            msg += " " + str(args[0])
            args = tuple(list(args)[1:])
        super().__init__(msg, *args, **kwargs)

    def make_msg(self, response: Response, details: str):
        servicename = response.service.get_name() + " " if response.service is not None else ""
        return "%sservice execution returned a valid response but unexpected. Details : %s " % (servicename, details)


class ConfigError(Exception):
    """
    Raised when a bad configuration element is encountered.

    :param key: The configuration key that failed to resolve properly
    :type key: object

    """
    key: Any

    def __init__(self, key: Any, msg="<No details given>", *args, **kwargs):
        self.key = key
        super().__init__(msg, *args, **kwargs)



from udsoncan import services
from udsoncan.exceptions import *
from udsoncan import Request, Response

from test.ClientServerTest import ClientServerTest
import time

class TestClient(ClientServerTest):
    def __init__(self, *args, **kwargs):
        ClientServerTest.__init__(self, *args, **kwargs)

    def test_timeout_override(self):
        pass

    def _test_timeout_override(self):
        req = Request(service = services.TesterPresent, subfunction=0) 
        timeout = 0.5
        try:
            t1 = time.time()
            response = self.udsclient.send_request(req, timeout=timeout)
            raise Exception('Request did not raise a TimeoutException')
        except TimeoutException as e:
            diff = time.time() - t1
            self.assertGreater(diff, timeout, 'Timeout raised after %.3f seconds when it should be %.3f sec' % (diff, timeout))
            self.assertLess(diff, timeout+0.5, 'Timeout raised after %.3f seconds when it should be %.3f sec' % (diff, timeout))

    # Server does not respond. Overall timeout is set smaller than P2 timeout. Overall timeout should trig
    def test_no_response_overall_timeout(self):
        pass

    def _test_no_response_overall_timeout(self):
        req = Request(service = services.TesterPresent, subfunction=0) 
        timeout = 0.5
        self.udsclient.set_configs({'request_timeout': timeout, 'p2_timeout' : 2, 'p2_star_timeout':2})
        try:
            t1 = time.time()
            response = self.udsclient.send_request(req)
            raise Exception('Request did not raise a TimeoutException')
        except TimeoutException as e:
            diff = time.time() - t1
            self.assertGreater(diff, timeout, 'Timeout raised after %.3f seconds when it should be %.3f sec' % (diff, timeout))
            self.assertLess(diff, timeout+0.5, 'Timeout raised after %.3f seconds when it should be %.3f sec' % (diff, timeout))

    # Server does not respond. P2 timeout is smaller than overall timeout. P2 timeout should trig
    def test_no_response_p2_timeout(self):
        pass

    def _test_no_response_p2_timeout(self):
        req = Request(service = services.TesterPresent, subfunction=0) 
        timeout = 0.5
        self.udsclient.set_configs({'request_timeout': 2, 'p2_timeout' : timeout, 'p2_star_timeout':2})
        try:
            t1 = time.time()
            response = self.udsclient.send_request(req)
            raise Exception('Request did not raise a TimeoutException')
        except TimeoutException as e:
            diff = time.time() - t1
            self.assertGreater(diff, timeout, 'Timeout raised after %.3f seconds when it should be %.3f sec' % (diff, timeout))
            self.assertLess(diff, timeout+0.5, 'Timeout raised after %.3f seconds when it should be %.3f sec' % (diff, timeout))

    # Overall timeout is set to 0.5. Server respond "pendingResponse" for 1 sec. 
    # Client should timeout first as we are setting respect_overall_timeout True.
    def test_overall_timeout_pending_response(self):
        if not hasattr(self, 'completed'):
            self.completed = False
        self.conn.touserqueue.get(timeout=0.2)
        response = Response(service=services.TesterPresent, code=Response.Code.RequestCorrectlyReceived_ResponsePending)
        t1 = time.time()
        while time.time() - t1 < 1 and not self.completed:
            self.conn.fromuserqueue.put(response.get_payload())
            time.sleep(0.1)

    def _test_overall_timeout_pending_response(self):
        req = Request(service = services.TesterPresent, subfunction=0) 
        timeout = 0.5
        self.udsclient.set_configs({'request_timeout': timeout, 'p2_timeout' : 2, 'p2_star_timeout':2})
        try:
            t1 = time.time()
            response = self.udsclient.send_request(req)
            raise Exception('Request did not raise a TimeoutException')
        except TimeoutException as e:
            self.assertIsNotNone(self.udsclient.last_response, 'Client never received the PendingResponse message')
            self.completed = True
            diff = time.time() - t1
            self.assertGreater(diff, timeout, 'Timeout raised after %.3f seconds when it should be %.3f sec' % (diff, timeout))
            self.assertLess(diff, timeout+0.5, 'Timeout raised after %.3f seconds when it should be %.3f sec' % (diff, timeout))

    #  Sends 2 "pending response" response to switch to P2* timeout.
    def test_p2_star_timeout(self):
        self.conn.touserqueue.get(timeout=0.2)
        response = Response(service=services.TesterPresent, code=Response.Code.RequestCorrectlyReceived_ResponsePending)
        self.conn.fromuserqueue.put(response.get_payload())
        time.sleep(0.1)
        self.conn.fromuserqueue.put(response.get_payload())

    def _test_p2_star_timeout(self):
        req = Request(service = services.TesterPresent, subfunction=0) 
        timeout = 2
        self.udsclient.set_configs({'request_timeout': 5, 'p2_timeout' : 0.5, 'p2_star_timeout':timeout})
        try:
            t1 = time.time()
            response = self.udsclient.send_request(req)
            raise Exception('Request did not raise a TimeoutException')
        except TimeoutException as e:
            self.assertIsNotNone(self.udsclient.last_response, 'Client never received the PendingResponse message')
            self.completed = True
            diff = time.time() - t1
            self.assertGreater(diff, timeout, 'Timeout raised after %.3f seconds when it should be %.3f sec' % (diff, timeout))
            self.assertLess(diff, timeout+0.5, 'Timeout raised after %.3f seconds when it should be %.3f sec' % (diff, timeout))

    # send RequestCorrectlyReceived_ResponsePending responses until well beyond overall_timeout to check response
    # This function is used in 2 tests:
    #  * checking there is a timeout when the overall_timeout is repected
    #  * checking that the full transaction is allowed to complete with the overall_timeout is ignored 
    def RCRRP_responses(self):
        self.conn.touserqueue.get(timeout=0.2)
        response = Response(service=services.TesterPresent, code=Response.Code.RequestCorrectlyReceived_ResponsePending)
        t1 = time.time()
        while time.time() - t1 <= 4.0:
            self.conn.fromuserqueue.put(response.get_payload())
            time.sleep(0.1)
        response = Response(service=services.TesterPresent, code=Response.Code.PositiveResponse, data=bytes(2))
        self.conn.fromuserqueue.put(response.get_payload())
    
    def RCRRP_responses_check(self, respect_overall_timeout):
        req = Request(service = services.TesterPresent, subfunction=0) 
        overall_timeout = 2.0
        if not respect_overall_timeout:
            overall_timeout = None
        p2_star_timeout = 1.0

        self.udsclient.set_configs({'request_timeout': overall_timeout, 'p2_timeout' : 0.5, 'p2_star_timeout':p2_star_timeout})

        # Record our expectation on how long the timeout wlil be
        if respect_overall_timeout:
            timeout = overall_timeout
        else:
            # Server is going to send RCRRP for 4s.  Then the client will wait for P2*
            timeout = 4.0 + p2_star_timeout

        t1 = time.time()
        try:
            response = self.udsclient.send_request(req)
        except TimeoutException as e:
            if respect_overall_timeout:
                diff = time.time() - t1
                self.assertGreater(diff, timeout, 'Timeout raised after %.3f seconds when it should be %.3f sec' % (diff, timeout))
                self.assertLess(diff, timeout+0.5, 'Timeout raised after %.3f seconds when it should be %.3f sec' % (diff, timeout))
                self.assertIsNotNone(self.udsclient.last_response, 'Client never received the PendingResponse message')
            else:
                raise Exception('Request raised a TimeoutException')
        if not respect_overall_timeout:
            self.assertEqual(self.udsclient.last_response.code, Response.Code.PositiveResponse, 'Client never received the Positive Response')

    #  Sends "pending response" responses to go past the overall_timeout and check there is a timeout when overall_timeout is respected.
    def test_RCRRP_with_overall_timeout(self):
        self.RCRRP_responses()

    def _test_RCRRP_with_overall_timeout(self):
        self.RCRRP_responses_check(respect_overall_timeout=True)

    # Sends "pending response" responses to got past the overall_timeout and check there is no timeout in default state.
    def test_RCRRP_no_overall_timeout(self):
        self.RCRRP_responses()

    def _test_RCRRP_no_overall_timeout(self):
        self.RCRRP_responses_check(respect_overall_timeout=False)

    #  Sends 2 "pending response" response to switch to P2* timeout.
    def test_p2_star_timeout_overrided_by_diagnostic_session_control(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x50\x01\x03\xE8\x00\xC8")  # Respond to diagnostic session control with timeout of 2 sec
        
        response = Response(service=services.TesterPresent, code=Response.Code.RequestCorrectlyReceived_ResponsePending)
        self.conn.fromuserqueue.put(response.get_payload())
        time.sleep(0.1)
        self.conn.fromuserqueue.put(response.get_payload())

    def _test_p2_star_timeout_overrided_by_diagnostic_session_control(self):
        req = Request(service = services.TesterPresent, subfunction=0) 
        self.udsclient.set_configs({'request_timeout': 5, 'p2_timeout' : 0.5, 'p2_star_timeout':1})
        self.udsclient.change_session(1)
        timeout = 2
        try:
            t1 = time.time()
            response = self.udsclient.send_request(req)
            raise Exception('Request did not raise a TimeoutException')
        except TimeoutException as e:
            self.assertIsNotNone(self.udsclient.last_response, 'Client never received the PendingResponse message')
            self.completed = True
            diff = time.time() - t1
            self.assertGreater(diff, timeout, 'Timeout raised after %.3f seconds when it should be %.3f sec' % (diff, timeout))
            self.assertLess(diff, timeout+0.5, 'Timeout raised after %.3f seconds when it should be %.3f sec' % (diff, timeout))

    def test_payload_override_literal(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b'\x12\x34\x56\x78')
        self.conn.fromuserqueue.put(b"\x7E\x00")

    def _test_payload_override_literal(self):
        req = Request(service = services.TesterPresent, subfunction=0) 
        with self.udsclient.payload_override(b'\x12\x34\x56\x78'):
            response = self.udsclient.send_request(req)
            self.assertEqual(response.original_payload, b'\x7E\x00')


    def test_payload_override_func(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b'\x99\x88\x77\x66')
        self.conn.fromuserqueue.put(b"\x7E\x00")

    def _test_payload_override_func(self):
        def func(payload):
            return b'\x99\x88\x77\x66'

        req = Request(service = services.TesterPresent, subfunction=0) 
        with self.udsclient.payload_override(func):
            response = self.udsclient.send_request(req)
            self.assertEqual(response.original_payload, b'\x7E\x00')